# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Ganache Docker Image."""
import logging
import time
from typing import Dict, List, Optional

import requests
from aea.exceptions import enforce
from aea_test_autonomy.docker.base import DockerImage
from docker import DockerClient
from docker.models.containers import Container


DEFAULT_GANACHE_ADDR = "http://127.0.0.1"
DEFAULT_GANACHE_PORT = 8545
DEFAULT_GANACHE_CHAIN_ID = 1337


class GanacheDockerImage(DockerImage):
    """Wrapper to Ganache Docker image."""

    def __init__(
        self,
        client: DockerClient,
        addr: str,
        port: int,
        config: Optional[Dict] = None,
        gas_limit: str = "0x9184e72a000",  # 10000000000000,
    ):
        """
        Initialize the Ganache Docker image.

        :param client: the Docker client.
        :param addr: the address.
        :param port: the port.
        :param config: optional configuration to command line.
        :param gas_limit: the gas limit for blocks.
        """
        super().__init__(client)
        self._addr = addr
        self._port = port
        self._config = config or {}
        self._gas_limit = gas_limit

    @property
    def tag(self) -> str:
        """Get the image tag."""
        return "trufflesuite/ganache:beta"

    def _make_ports(self) -> Dict:
        """Make ports dictionary for Docker."""
        return {f"{self._port}/tcp": ("0.0.0.0", self._port)}  # nosec

    def _build_command(self) -> List[str]:
        """Build command."""
        # cmd = ["--chain.hardfork=london"]  # noqa: E800
        cmd = ["--miner.blockGasLimit=" + str(self._gas_limit)]
        # cmd += ["--miner.callGasLimit=" + "0x1fffffffffffff"]  # noqa: E800
        accounts_balances = self._config.get("accounts_balances", [])
        for account, balance in accounts_balances:
            cmd += [f"--wallet.accounts='{account},{balance}'"]
        return cmd

    def create(self) -> Container:
        """Create the container."""
        cmd = self._build_command()
        container = self._client.containers.run(
            self.tag, command=cmd, detach=True, ports=self._make_ports()
        )
        return container

    def create_many(self, nb_containers: int) -> List[Container]:
        """Instantiate the image in many containers, parametrized."""
        raise NotImplementedError()

    def wait(self, max_attempts: int = 15, sleep_rate: float = 1.0) -> bool:
        """Wait until the image is up."""
        request = dict(jsonrpc=2.0, method="web3_clientVersion", params=[], id=1)
        for i in range(max_attempts):
            try:
                response = requests.post(f"{self._addr}:{self._port}", json=request)
                enforce(response.status_code == 200, "")
                return True
            except Exception:  # pylint: disable=broad-except
                logging.error(
                    "Attempt %s failed. Retrying in %s seconds...", i, sleep_rate
                )
                time.sleep(sleep_rate)
        return False


class GanacheForkDockerImage(GanacheDockerImage):
    """Extends GanacheDockerImage to enable forking"""

    NETWORK: str = "ropsten"
    BLOCK_NUMBER: int = 11290000

    def _build_command(self) -> List[str]:
        """Build command."""

        cmd = [
            "--wallet.deterministic=true",
            f"--fork.network={self.NETWORK}",
            f"--fork.blockNumber={self.BLOCK_NUMBER}",
        ]

        return cmd



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Utilities for the Open Autonomy test tools."""
import contextlib
import logging
import os
import time
from os import PathLike
from typing import Any, Generator

import requests
from aea_test_autonomy.configurations import DEFAULT_REQUESTS_TIMEOUT, MAX_RETRIES


def tendermint_health_check(
    url: str,
    max_retries: int = MAX_RETRIES,
    sleep_interval: float = 1.0,
    timeout: float = DEFAULT_REQUESTS_TIMEOUT,
) -> bool:
    """Wait until a Tendermint RPC server is up."""
    attempt = 0
    while attempt < max_retries:
        try:
            response = requests.get(url + "/health", timeout=timeout)
            assert response.status_code == 200  # nosec
            return True
        except (AssertionError, requests.exceptions.ConnectionError):
            attempt += 1
        logging.debug(
            f"Health-check attempt {attempt - 1} for {url} failed. Retrying in {sleep_interval} seconds..."
        )
        time.sleep(sleep_interval)
    return False


@contextlib.contextmanager
def cd(path: PathLike) -> Generator:  # pragma: nocover
    """Change working directory temporarily."""
    old_path = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(old_path)


def try_send(gen: Generator, obj: Any = None) -> None:
    """
    Try to send an object to a generator.

    :param gen: the generator.
    :param obj: the object.
    """
    with contextlib.suppress(StopIteration):
        gen.send(obj)


def identity(arg: Any) -> Any:
    """Define an identity function."""
    return arg



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""ACN Docker Image."""
import logging
import socket
import time
from typing import Dict, List, Optional

from aea.exceptions import enforce
from aea_test_autonomy.docker.base import DockerImage
from docker import DockerClient
from docker.models.containers import Container


_LOCAL_ADDRESS = "0.0.0.0"  # nosec

DEFAULT_ACN_CONFIG: Dict[str, str] = dict(
    AEA_P2P_ID="d9e43d3f0266d14b3af8627a626fa734450b1c0fcdec6f88f79bcf5543b4668c",
    AEA_P2P_URI_PUBLIC=f"{_LOCAL_ADDRESS}:5000",
    AEA_P2P_URI=f"{_LOCAL_ADDRESS}:5000",
    AEA_P2P_DELEGATE_URI=f"{_LOCAL_ADDRESS}:11000",
    AEA_P2P_URI_MONITORING=f"{_LOCAL_ADDRESS}:8081",
    ACN_LOG_FILE="/acn/libp2p_node.log",
)


class ACNNodeDockerImage(DockerImage):
    """Wrapper to ACNNode Docker image."""

    uris: List = [
        "AEA_P2P_URI_PUBLIC",
        "AEA_P2P_URI",
        "AEA_P2P_DELEGATE_URI",
        "AEA_P2P_URI_MONITORING",
    ]

    def __init__(
        self,
        client: DockerClient,
        config: Optional[Dict] = None,
    ):
        """
        Initialize the ACNNode Docker image.

        :param client: the Docker client.
        :param config: optional configuration to command line.
        """
        super().__init__(client)
        self._config = config or {}

    @property
    def tag(self) -> str:
        """Get the image tag."""
        return "valory/acn-node:latest"

    def _make_ports(self) -> Dict:
        """Make ports dictionary for Docker."""

        return {
            f"{port}/tcp": (_LOCAL_ADDRESS, port)
            for port in [self._config[uri].split(":")[1] for uri in self.uris]
        }

    def create(self) -> Container:
        """Create the container."""
        container = self._client.containers.run(
            self.tag,
            "--config-from-env",
            detach=True,
            ports=self._make_ports(),
            environment=self._config,
        )
        return container

    def create_many(self, nb_containers: int) -> List[Container]:
        """Instantiate the image in many containers, parametrized."""
        raise NotImplementedError()

    def wait(self, max_attempts: int = 15, sleep_rate: float = 1.0) -> bool:
        """Wait until the image is up."""

        i, to_be_connected = 0, {self._config[uri] for uri in self.uris}
        while i < max_attempts and to_be_connected:
            i += 1
            for uri in to_be_connected:
                try:
                    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                    host, port = uri.split(":")
                    result = sock.connect_ex((host, int(port)))
                    sock.close()
                    enforce(result == 0, "")
                    to_be_connected.remove(uri)
                    logging.info(f"URI ready: {uri}")
                    break
                except Exception:  # pylint: disable=broad-except
                    logging.error(
                        f"Attempt {i} failed on {uri}. Retrying in {sleep_rate} seconds..."
                    )
                    time.sleep(sleep_rate)

        return not to_be_connected
